// Cone.cpp: Implementation of the Cone class.
//			 Parent Class: Object
//			 This class stores the representation of a Cone object	
//			 and provides a method to intersect a ray with the cone
//////////////////////////////////////////////////////////////////////

#include "Cone.h"
#include "Misc.h"

#define WALL 0
#define BASE 1
#define CAP 2

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Cone::Cone()
{
	this->radius = 1;
	this->height = 1;
}

Cone::~Cone()
{
}

Cone::Cone(double rad, double height)
{
	this->radius = rad;
	this->height = height;
}
	
void Cone::Set(double rad, double height)
{
	this->radius = rad;
	this->height = height;
}

//////////////////////////////////////////////////////////////////////
// Intersect : Tests for intersection between the given ray.  If there
// is an intersection, the hit information is stored in ho, and true
// is returned, otherwise false is returned.
//////////////////////////////////////////////////////////////////////
bool Cone::Intersect(Ray ray, HitObject& ho)
{
	double a,b,c,d;
	bool conehit1 = false, conehit2 = false;
	bool caphit1 = false, conehit = false, caphit = false;
  
	double x1,y1,z1,x2,y2,z2,xcyl,ycyl,zcyl,xcap,ycap,zcap;
	double tcap,tcyl,t1,t2;
	double dx,dz, dy;
	
	Vec3 normal;

	// transform the ray according to this objects transformation
	// matrix
	Ray newRay = ray.InverseTransRay(this->transMatrix);
	
	double vdx = newRay.v3Dir.x;
	double vdy = newRay.v3Dir.y;
	double vdz = newRay.v3Dir.z;
	double vox = newRay.v4Ori.x;
	double voy = newRay.v4Ori.y;
	double voz = newRay.v4Ori.z;
 
	double vdxsq = vdx*vdx;
	double vdysq = vdy*vdy;
	double vdzsq = vdz*vdz;
	double voxsq = vox*vox;
	double voysq = voy*voy;
	double vozsq = voz*voz;
 
	double r2divh2 = (radius * radius) / (height * height);
	double h = height;
	double hsq = h * h;
	double r = radius;
	
	// Get the quadratic for the cone body:
	// a t^2 + b t + c = 0
	
	a = vdxsq + vdysq - (r2divh2 * vdzsq);

	b = (2*vox*vdx) + (2*voy*vdy) - (r2divh2*2*voz*vdz) + (r2divh2*2*vdz*h);

	c = voxsq + voysq - (r2divh2 * vozsq) + (r2divh2 * 2 * voz * h) - (r2divh2 * hsq);

	// discriminant
	d = sqr(b) - 4*a*c;
	 

	if (d < 0.0) 
	{
		// no hit at all
	}
	else 
	{		
		t1 = (-b - sqrt(d))/(2*a);
		
		if (t1 > EPSILON) 
		{    

			x1 = newRay.v4Ori.x + t1 * newRay.v3Dir.x;
			y1 = newRay.v4Ori.y + t1 * newRay.v3Dir.y;
			z1 = newRay.v4Ori.z + t1 * newRay.v3Dir.z;

			if (0 <= z1 && z1 <= height) 
				conehit1 = true;
		}

		t2 = (-b + sqrt(d))/(2*a);
			
		if (t2 > EPSILON)
		{    
			x2 = newRay.v4Ori.x + t2 * newRay.v3Dir.x;
			y2 = newRay.v4Ori.y + t2 * newRay.v3Dir.y;
			z2 = newRay.v4Ori.z + t2 * newRay.v3Dir.z;

			if (0 <= z2 && z2 <= height) 
				conehit2 = true;			
		}

		if ((conehit1 && conehit2 && t1 < t2) ||
			(conehit1 && !conehit2)) 
		{
			xcyl = x1;
			ycyl = y1;
			zcyl = z1;
			tcyl = t1;
			conehit = true;
		}
		else if ((conehit1 && conehit2 && t1 > t2) ||
			    (!conehit1 && conehit2)) 
		{
			xcyl = x2;
			ycyl = y2;
			zcyl = z2;
			tcyl = t2;
			conehit = true;
		}
		else 
		{
			conehit = false;
		}
	}

	// Now, get the hitpoints on the end caps, if any.
	
	caphit = false;
	
	t1 = (-newRay.v4Ori.z) / newRay.v3Dir.z;
	if (t1 > 1) 
	{    
	    x2 = newRay.v4Ori.x + t1 * newRay.v3Dir.x;
	    y2 = newRay.v4Ori.y + t1 * newRay.v3Dir.y;
	    z2 = newRay.v4Ori.z + t1 * newRay.v3Dir.z;
    
	    dx = x2;
		dy = y2;

		if (dx*dx + dy*dy < radius*radius) 
		{
			caphit = true;
		}
	}

	if (caphit) 
	{
		xcap = x1;
		ycap = y1;
		zcap = z1;
		tcap = t1;
	    caphit = true;
	}
	else 
	{
		caphit = false;
	}
	
  
	// if there were any hits, get the one with 
	// the minimum t value, calculate it's normal,
	// store this info in a hit object and return 
	// true, otherwise return false if there were
	// no hits.
	if ((conehit && caphit && tcyl < tcap) ||
		(conehit && !caphit)) 
	{
		Vec4 hitpoint;
	  
		hitpoint.Set(xcyl, ycyl, zcyl, 1);

		normal = Vec3( hitpoint.x,
					   hitpoint.y,
					   height - hitpoint.z);

		ho.Set(ray, tcyl, normal);

		return true;
	}
	else if ((conehit && caphit && tcyl > tcap) ||
			(!conehit && caphit)) 
	{

		normal.Set(0,0,-1);

		ho.Set(ray, tcap, normal);

	    return true;
	}
	else
	    return false;
}
