// Matrix.cpp: implementation of the Matrix class.
//
//////////////////////////////////////////////////////////////////////
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "Matrix.h"
#include "Misc.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
Matrix::Matrix()
{
	int i, j;
	
	for (i = 0; i<4; i++)
		for (j = 0; j<4; j++)
			mat[i][j] = 0.0f;
	

	mat[0][0] = mat[1][1] = mat[2][2] = mat[3][3] = 1.0f;

	invert();
}

Matrix::~Matrix()
{
}

void Matrix::setIdentity()
{
	int i, j;

	for (i = 0; i<4; i++)
		for (j = 0; j<4; j++)
			mat[i][j] = 0.0f;
	

	mat[0][0] = mat[1][1] = mat[2][2] = mat[3][3] = 1.0f;

	invert();
}

//////////////////////////////////////////////////////////////////////
// scale : Scales by sx, sy, sz
//////////////////////////////////////////////////////////////////////
void Matrix::scale(double sx, double sy, double sz)
{
	if (abs((int)sx) < EPSILON || abs((int)sy) < EPSILON || abs((int)sz) < EPSILON)
		return;

	double smat[4][4];

	int i, j;
	
	for (i = 0; i<4; i++)
		for (j = 0; j<4; j++)
			smat[i][j] = 0.0f;

	smat[0][0] = smat[1][1] = smat[2][2] = smat[3][3] = 1.0f;

	smat[0][0] = sx;
	smat[1][1] = sy;
	smat[2][2] = sz;

	// add to our current trasnformation matrix
	MultMat(smat);
}

//////////////////////////////////////////////////////////////////////
// translate : Translates by tx, ty, tz
//////////////////////////////////////////////////////////////////////
void Matrix::translate(double tx, double ty, double tz)
{
	mat[0][3] = mat[0][3] + tx;
	mat[1][3] = mat[1][3] + ty;
	mat[2][3] = mat[2][3] + tz;

	// recompute the inverse and transpose inverse
	invert();
}

//////////////////////////////////////////////////////////////////////
// MultMat : Multiplies Mat by m2 (4x4)
//////////////////////////////////////////////////////////////////////
void Matrix::MultMat(double m2[4][4])
{
	int i,j;
	double newMat[4][4];

	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			newMat[i][j] = mat[i][0]*m2[0][j] + mat[i][1]*m2[1][j] + mat[i][2]*m2[2][j] + mat[i][3]*m2[3][j];
	
	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			mat[i][j] = newMat[i][j];

	// recompute the inverse and transpose inverse
	invert();
}

//////////////////////////////////////////////////////////////////////
// invert : Calculates the invese and inverse transpose of mat and
//			stores them in their own double arrays.
//////////////////////////////////////////////////////////////////////
void Matrix::invert() {
	int i, j;
	double det;
	double tmp;

	adjoint();
	
	det = det4x4( mat );

	if ( fabs( det ) < EPSILON) {
		fprintf(stderr,"invert_matrix: matrix is singular!");
		return;
	}

	// calculate inverse
	for (i=0; i<4; i++)
		for(j=0; j<4; j++)
		{
			inv[i][j] = inv[i][j] / det;
			invTrans[i][j] = inv[i][j];
		}

	// get transpose now
	for (i=0; i<4; i++)
		for(j=i; j<4; j++)
		{
			tmp = invTrans[i][j];
			invTrans[i][j] = invTrans[j][i];
			invTrans[j][i] = tmp;
		}
}

//////////////////////////////////////////////////////////////////////
// Adjoint
//////////////////////////////////////////////////////////////////////
void Matrix::adjoint() {
  double a1, a2, a3, a4, b1, b2, b3, b4;
  double c1, c2, c3, c4, d1, d2, d3, d4;
  

  a1 = mat[0][0]; b1 = mat[0][1]; 
  c1 = mat[0][2]; d1 = mat[0][3];

  a2 = mat[1][0]; b2 = mat[1][1]; 
  c2 = mat[1][2]; d2 = mat[1][3];

  a3 = mat[2][0]; b3 = mat[2][1];
  c3 = mat[2][2]; d3 = mat[2][3];

  a4 = mat[3][0]; b4 = mat[3][1]; 
  c4 = mat[3][2]; d4 = mat[3][3];

  inv[0][0]  =   det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4);
  inv[1][0]  = - det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4);
  inv[2][0]  =   det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4);
  inv[3][0]  = - det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);
        
  inv[0][1]  = - det3x3( b1, b3, b4, c1, c3, c4, d1, d3, d4);
  inv[1][1]  =   det3x3( a1, a3, a4, c1, c3, c4, d1, d3, d4);
  inv[2][1]  = - det3x3( a1, a3, a4, b1, b3, b4, d1, d3, d4);
  inv[3][1]  =   det3x3( a1, a3, a4, b1, b3, b4, c1, c3, c4);
        
  inv[0][2]  =   det3x3( b1, b2, b4, c1, c2, c4, d1, d2, d4);
  inv[1][2]  = - det3x3( a1, a2, a4, c1, c2, c4, d1, d2, d4);
  inv[2][2]  =   det3x3( a1, a2, a4, b1, b2, b4, d1, d2, d4);
  inv[3][2]  = - det3x3( a1, a2, a4, b1, b2, b4, c1, c2, c4);
        
  inv[0][3]  = - det3x3( b1, b2, b3, c1, c2, c3, d1, d2, d3);
  inv[1][3]  =   det3x3( a1, a2, a3, c1, c2, c3, d1, d2, d3);
  inv[2][3]  = - det3x3( a1, a2, a3, b1, b2, b3, d1, d2, d3);
  inv[3][3]  =   det3x3( a1, a2, a3, b1, b2, b3, c1, c2, c3);
}

//////////////////////////////////////////////////////////////////////
// det4x4 : 4x4 determinant
//////////////////////////////////////////////////////////////////////
double Matrix::det4x4( double m[4][4] ) {
  double ans;
  double a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4;

  a1 = m[0][0]; b1 = m[0][1]; 
  c1 = m[0][2]; d1 = m[0][3];

  a2 = m[1][0]; b2 = m[1][1]; 
  c2 = m[1][2]; d2 = m[1][3];

  a3 = m[2][0]; b3 = m[2][1]; 
  c3 = m[2][2]; d3 = m[2][3];

  a4 = m[3][0]; b4 = m[3][1]; 
  c4 = m[3][2]; d4 = m[3][3];

  ans = a1 * det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4)
    - b1 * det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4)
    + c1 * det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4)
    - d1 * det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);
  return ans;
}

//////////////////////////////////////////////////////////////////////
// det3x3 : 3x3 determinant
//////////////////////////////////////////////////////////////////////
double Matrix::det3x3( double a1, double a2, double a3, double b1, double b2, double b3,double c1, double c2, double c3 )
{
  double ans;

  ans = a1 * det2x2( b2, b3, c2, c3 )
    - b1 * det2x2( a2, a3, c2, c3 )
    + c1 * det2x2( a2, a3, b2, b3 );
  return ans;
}

//////////////////////////////////////////////////////////////////////
// det2x2 : 2x2 determinant
//////////////////////////////////////////////////////////////////////
double Matrix::det2x2( double a, double b, double c, double d)
{
  double ans;
  ans = a * d - b * c;
  return ans;
}

//////////////////////////////////////////////////////////////////////
// MultVector3ByInverse : multiples the inverse matrix by 
// the passed in Vec3 Vector.  Useful for transforming a ray.
//////////////////////////////////////////////////////////////////////
Vec3 Matrix::MultVector3ByInverse (const Vec3 V) const
{
	Vec3 Vnew;
	
	Vnew.x = inv[0][0] * V.x + inv[0][1] * V.y + inv[0][2] * V.z;
	Vnew.y = inv[1][0] * V.x + inv[1][1] * V.y + inv[1][2] * V.z;
	Vnew.z = inv[2][0] * V.x + inv[2][1] * V.y + inv[2][2] * V.z;

	return Vnew;
}

//////////////////////////////////////////////////////////////////////
// MultVector4ByInverse : multiples the inverse matrix by 
// the passed in Vec4 Vector.  Useful for transforming a ray.
//////////////////////////////////////////////////////////////////////
Vec4 Matrix::MultVector4ByInverse (const Vec4 V)
{
	Vec4 Vnew;
	
	Vnew.x = inv[0][0] * V.x + inv[0][1] * V.y + inv[0][2] * V.z + inv[0][3] * 1;
	Vnew.y = inv[1][0] * V.x + inv[1][1] * V.y + inv[1][2] * V.z + inv[1][3] * 1;
	Vnew.z = inv[2][0] * V.x + inv[2][1] * V.y + inv[2][2] * V.z + inv[2][3] * 1;
	Vnew.w = inv[3][0] * V.x + inv[3][1] * V.y + inv[3][2] * V.z + inv[3][3] * 1;

	return Vnew;
}

//////////////////////////////////////////////////////////////////////
// MultVector3ByInverseTranspose : multiples the inverse transpose 
// matrix by the passed in Vec3 Vector.  Useful for transforming 
// the normal.
//////////////////////////////////////////////////////////////////////
Vec3 Matrix::MultVector3ByInverseTranspose (const Vec3 V) const
{
	Vec3 Vnew;
	
	Vnew.x = invTrans[0][0] * V.x + invTrans[0][1] * V.y + invTrans[0][2] * V.z;
	Vnew.y = invTrans[1][0] * V.x + invTrans[1][1] * V.y + invTrans[1][2] * V.z;
	Vnew.z = invTrans[2][0] * V.x + invTrans[2][1] * V.y + invTrans[2][2] * V.z;

	return Vnew;
}

//////////////////////////////////////////////////////////////////////
// rotateX : rotates by angle in the X direction
//////////////////////////////////////////////////////////////////////
void Matrix::rotatex(double angle)
{
	double rotmat[4][4];

	int i, j;
	
	for (i = 0; i<4; i++)
		for (j = 0; j<4; j++)
			rotmat[i][j] = 0.0f;


	rotmat[0][0] = rotmat[1][1] = rotmat[2][2] = rotmat[3][3] = 1.0f;

	double ang = angle * PI/180;
	double c = (cos(ang)), s = sin(ang);

	rotmat[1][1] = c;
	rotmat[1][2] = -s;
	rotmat[2][1] = s;
	rotmat[2][2] = c;

	// add to our current trasnformation matrix
	MultMat(rotmat);
}

//////////////////////////////////////////////////////////////////////
// rotateY : rotates by angle in the Y direction
//////////////////////////////////////////////////////////////////////
void Matrix::rotatey(double angle)
{
	double rotmat[4][4];

	int i, j;
	
	for (i = 0; i<4; i++)
		for (j = 0; j<4; j++)
			rotmat[i][j] = 0.0f;


	rotmat[0][0] = rotmat[1][1] = rotmat[2][2] = rotmat[3][3] = 1.0f;

	double ang = angle * PI/180;
	double c = (cos(ang)), s = sin(ang);

	rotmat[0][0] = c;
	rotmat[0][2] = s;
	rotmat[2][0] = -s;
	rotmat[2][2] = c;

	// add to our current trasnformation matrix
	MultMat(rotmat);
}

//////////////////////////////////////////////////////////////////////
// rotateZ : rotates by angle in the Z direction
//////////////////////////////////////////////////////////////////////
void Matrix::rotatez(double angle)
{
	double rotmat[4][4];

	int i, j;
	
	for (i = 0; i<4; i++)
		for (j = 0; j<4; j++)
			rotmat[i][j] = 0.0f;


	rotmat[0][0] = rotmat[1][1] = rotmat[2][2] = rotmat[3][3] = 1.0f;

	double ang = angle * PI/180;
	double c = (cos(ang)), s = sin(ang);

	rotmat[0][0] = c;
	rotmat[0][1] = -s;
	rotmat[1][0] = s;
	rotmat[1][1] = c;

	// add to our current trasnformation matrix
	MultMat(rotmat);
}

