// RayTracer.cpp: Implementation of the RayTracer class.
//				This class ties everything together.  It allows for
//				the addition of objects and lights to the scene. It
//				casts and tracks the rays and generates the final
//				array of pixels to display on screen.
//////////////////////////////////////////////////////////////////////

#include <string.h>
#include <assert.h>
#include "RayTracer.h"
#include "SurfaceModel.h"
#include "Misc.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RayTracer::RayTracer()
{
	img = NULL;

	xMin = 0.0;
	xMax = 1.0;
	yMin = 0.0;
	yMax = 1.0;
	Height = 400;
	Width = 400;

	img = new byte[Width*Height*3];

	BackgroundColor.Set(0,0,0);
	
	// Surface Properties
	SurfaceAmbient.Set(0,0,0);
	SurfaceDiffuse.Set(0,0,0);
	SurfaceReflective = 0;
	SurfaceTransparent = 0;
	SurfaceRefractionIndex = 1;
    SurfaceSpecExponent = 1;

	Height = 400;
	Width = 400;
    
    Depth = 3;
    Kd = 1;
    Ks = 1;
    Kt = 1;
}

RayTracer::~RayTracer()
{
}

//////////////////////////////////////////////////////////////////////
// Set Background Colour
//////////////////////////////////////////////////////////////////////
void RayTracer::Background(double r, double g, double b)
{
	BackgroundColor.Set((double)r, (double)g, (double)b);
}

//////////////////////////////////////////////////////////////////////
// Add a light with 0 radius
//////////////////////////////////////////////////////////////////////
void RayTracer::AddLight(double intensity, double x, double y, double z)
{
    Lights.Add(intensity, x, y, z, 0);
}

//////////////////////////////////////////////////////////////////////
// Add a light with radiux (NOT IMPLEMENTED)
//////////////////////////////////////////////////////////////////////
void RayTracer::AddLight(double intensity, double x, double y, double z, double R)
{
    Lights.Add(intensity, x, y, z, R);
}

//////////////////////////////////////////////////////////////////////
// Set Ambient surface properties
//////////////////////////////////////////////////////////////////////
void RayTracer::Ambient(double r, double g, double b)
{
	SurfaceAmbient.Set((double)r, (double)g, (double)b);
    SurfaceAmbient.Bind();
}

//////////////////////////////////////////////////////////////////////
// Set Diffuse surface properties
//////////////////////////////////////////////////////////////////////
void RayTracer::Diffuse(double r, double g, double b)
{
	SurfaceDiffuse.Set((double)r, (double)g, (double)b);
    SurfaceDiffuse.Bind();
}

//////////////////////////////////////////////////////////////////////
// Set Specular surface properties
//////////////////////////////////////////////////////////////////////
void RayTracer::Specular(double r, double g, double b)
{
	SurfaceSpecular.Set((double)r, (double)g, (double)b);
    SurfaceSpecular.Bind();
}

//////////////////////////////////////////////////////////////////////
// Set specular exponent property
//////////////////////////////////////////////////////////////////////
void RayTracer::SpecExponent(double exp)
{
    SurfaceSpecExponent = exp;
}

//////////////////////////////////////////////////////////////////////
// Sets Reflectivity index
//////////////////////////////////////////////////////////////////////
void RayTracer::Reflective(double reflect)
{
    assert( reflect>=0 && reflect<=1 );
	SurfaceReflective = reflect;
}

//////////////////////////////////////////////////////////////////////
// Sets Transparency index
//////////////////////////////////////////////////////////////////////
void RayTracer::Transparent(int trans)
{
    assert(trans>=0 && trans<=1 );
	SurfaceTransparent = trans;
}

//////////////////////////////////////////////////////////////////////
// Set index of refraction
//////////////////////////////////////////////////////////////////////
void RayTracer::Index(double index)
{
	SurfaceRefractionIndex = (double)index;
}

//////////////////////////////////////////////////////////////////////
// Adds a polygon described by the number of points in it, 
// and coordinates of those points
//////////////////////////////////////////////////////////////////////
void RayTracer::Polygon(int n, Vec4 *P)
{
	SurfaceModel surface(SurfaceAmbient, SurfaceDiffuse, SurfaceSpecular,
		SurfaceReflective, SurfaceTransparent, SurfaceRefractionIndex, SurfaceSpecExponent);

	ObjList.AddPolygon(n, P, surface);
}

//////////////////////////////////////////////////////////////////////
// Adds a cylinder described by a radius and height
//////////////////////////////////////////////////////////////////////
void RayTracer::Cylinder(double r, double height)
{
	SurfaceModel surface(SurfaceAmbient, SurfaceDiffuse, SurfaceSpecular,
		SurfaceReflective, SurfaceTransparent, SurfaceRefractionIndex, SurfaceSpecExponent);

	ObjList.AddCylinder(r, height, surface);
}

//////////////////////////////////////////////////////////////////////
// Adds a cone described by a radius and height
//////////////////////////////////////////////////////////////////////
void RayTracer::Cone(double r, double height)
{
	SurfaceModel surface(SurfaceAmbient, SurfaceDiffuse, SurfaceSpecular,
		SurfaceReflective, SurfaceTransparent, SurfaceRefractionIndex, SurfaceSpecExponent);

	ObjList.AddCone(r, height, surface);
}

//////////////////////////////////////////////////////////////////////
// Adds a sphere described by a radius and center
//////////////////////////////////////////////////////////////////////
void RayTracer::Sphere(double radius)
{
	SurfaceModel surface(SurfaceAmbient, SurfaceDiffuse, SurfaceSpecular,
		SurfaceReflective, SurfaceTransparent, SurfaceRefractionIndex, SurfaceSpecExponent);

	ObjList.AddSphere(radius, surface );
}

//////////////////////////////////////////////////////////////////////
// scales the last object added to the scene by sx,sy,sz
//////////////////////////////////////////////////////////////////////
void RayTracer::scale(double sx, double sy, double sz)
{
	ObjList.scale(sx,sy,sz);
}

//////////////////////////////////////////////////////////////////////
// translates the last object added to the scene by tx,ty,tz
//////////////////////////////////////////////////////////////////////
void RayTracer::translate(double tx, double ty, double tz)
{
	ObjList.translate(tx,ty,tz);
}

//////////////////////////////////////////////////////////////////////
// rotates the last object added to the scene by angle in the 
// x directions
//////////////////////////////////////////////////////////////////////
void RayTracer::rotateX(double angle)
{
	ObjList.rotateX(angle);
}

//////////////////////////////////////////////////////////////////////
// rotates the last object added to the scene by angle in the 
// y directions
//////////////////////////////////////////////////////////////////////
void RayTracer::rotateY(double angle)
{
	ObjList.rotateY(angle);
}

//////////////////////////////////////////////////////////////////////
// rotates the last object added to the scene by angle in the 
// z directions
//////////////////////////////////////////////////////////////////////
void RayTracer::rotateZ(double angle)
{
	ObjList.rotateZ(angle);
}

//////////////////////////////////////////////////////////////////////
// CastRays : Casts the rays from a static eyepoint and direction
// and calls TraceRay which traces the ray.  Each pixel is set to 
// the colour returned by TraceRay.
//////////////////////////////////////////////////////////////////////
void RayTracer::CastRays()
{
    int i, j;
	int index;

    for( i=-Height/2; i < Height/2; i++)
    {
        for( j=-Width/2; j < Width/2; j++)
        {
            Vec4 p;
            Vec3 d;
            Ray r;
            Colour color;

            // set eyepoint to a static location 
			// (0,0, -256);
			Vec3 dir ( (double)j, (double)i, (double)256 );

			// set direction to (0,0,256)
			Vec4 ori ( 0, 0, (double)-256, 1 );
			
			dir.Normalize();

			r.Set(ori, dir);

			index = ((Height-1-(i + (Height/2)))*Width + (j + (Width/2))) * 3;

			// Trace Ray
            color = TraceRay(r, Depth, -1);
			color.Bind();

			// set img
			img[index++] = (byte)(255*color.R);
			img[index++] = (byte)(255*color.G);
			img[index] = (byte)(255*color.B);
        }
    }
}

//////////////////////////////////////////////////////////////////////
// TraceRay : Tracks the rays progress through the scene.
// Spawns additional rays for reflection and refraction
//////////////////////////////////////////////////////////////////////
Colour RayTracer::TraceRay(Ray r, int depth, const int trans)
{
	Colour colour = BackgroundColor;
	HitObject ho;
	int which=-1;

    if( depth == 0 )
        return Colour(0,0,0);

	if( ObjList.Intersects(r, ho, which, trans) )
	{
        int i;
        Ray R, P;
        Vec3 dir, VV, normal, tmp;
        double kf, kn, reflect, transp;

		// start with ambient contribution
        colour = ObjList.GetAmbient(which);

		// calculate shading
        for( i=0; i<Lights.NumLights; i++ )
        {	
            Light light = Lights.Lights[i];
            Colour cd = ObjList.Shade(r, which, ho, light); 
            cd.Bind();
		    colour += Kd*cd; 
        }

        // compute Reflection & Refraction
		// refraction and reflection code is taken from: 
		// http://www.cs.unc.edu/~lee/class/comp238/hw2/
		dir = r.v3Dir;
        dir.Normalize();
		
		// transform the normal
        normal = ho.TransformNormal(ObjList.GetMatrix());
		normal.Normalize();

        VV = dir / fabs(dir * normal);
        tmp = VV + 2 * normal;
        tmp.Normalize();
		
		// set the refelection ray R and refractive ray P
        R.Set(ho.hit, tmp);
        kn = ObjList.GetRefractionIndex(which);

        kf = ( 1.0/sqrt( sqr(kn) * sqr(VV.Length() ) - sqr( (VV+normal).Length() ) ) );
        tmp = kf * (normal + VV) - normal;
        tmp.Normalize();
        P.Set(ho.hit, tmp);

        reflect = ObjList.IsReflective(which);
        transp = ObjList.IsTransparent(which);

		// Tracing the reflective ray
        if( !R.NoRay() && reflect > 0 && R.v3Dir * normal > 0)
        {
			R.v3Dir = -R.v3Dir;

            Colour cs = TraceRay(R, depth-1, which);
            cs.Bind();
            
            cs = cs * ObjList.GetSpecular(which);
			colour += Ks * reflect * cs;
        }
		
		// Tracing the transparency ray
        if( !P.NoRay() && transp > 0 )
        {
            Colour ct = TraceRay(P, depth-1, which);
            ct.Bind();
            colour += Kt * transp * ct;
        }

        colour.Bind();	
	}

	return colour;
}

//////////////////////////////////////////////////////////////////////
// Test : creates one of three scenes and renders the scene
// using ray tracing.
//////////////////////////////////////////////////////////////////////
void RayTracer::Test(int arg, int d)
{
	Depth = d;

	switch (arg)
	{
		case 0:

		Background(0, 0, 0);
		
		AddLight(1.5f, 600, 600, -750);
		
		// Polygon
		Ambient(0.2, 0.2, 0.2);
		Diffuse(0.2, 0.3, 0.2);
		Specular(0.628281, 0.555802, 0.866065);
    
		Reflective(1);
		SpecExponent(89);
		Transparent(0);
		
		Vec4 *P;
		P = new Vec4[4];

		P[0].x = -200;
		P[0].y = -200;
		P[0].z = 0;
		P[0].w = 1;

		P[1].x = -200;
		P[1].y = 200;
		P[1].z = 0;
		P[1].w = 1;

		P[2].x = 200;
		P[2].y = 200;
		P[2].z = 0;
		P[2].w = 1;

		P[3].x = 200;
		P[3].y = -200;
		P[3].z = 0;
		P[3].w = 1;

		Polygon(4, P);
		translate(0, 0, 200);

		// Top Plane
		Ambient(0.1, 0.1, 0.1);
		Diffuse(0.2, 0.3, 0.6);
		Specular(0.628281, 0.955802, 0.866065);
    
		Reflective(1);
		SpecExponent(44);
		Transparent(0);
		
		Vec4 *P2;
		P2 = new Vec4[4];

		P2[0].x = -200;
		P2[0].y = -100;
		P2[0].z = 0;
		P2[0].w = 1;

		P2[1].x = -200;
		P2[1].y = 100;
		P2[1].z = 0;
		P2[1].w = 1;

		P2[2].x = 200;
		P2[2].y = 100;
		P2[2].z = 0;
		P2[2].w = 1;

		P2[3].x = 200;
		P2[3].y = -100;
		P2[3].z = 0;
		P2[3].w = 1;

		Polygon(4, P2);

		rotateX(45);
		translate(0, -90, 200);

		// Middle Sphere
		Ambient(0.23125, 0.23125, 0.23125);
		Diffuse(0.2775, 0.2775, 0.2775);
		Specular(0.77391, 0.77391, 0.77391);
    
		Reflective(1);
		SpecExponent(33);
		Transparent(0);
		
		Sphere(50);
		translate(0, 100, 0);

		// Right Sphere
		Ambient(0.24725, 0.1995, 0.0745);
		Diffuse(0.75164, 0.60648, 0.22648);
		Specular(0.628281, 0.555802, 0.366065);
    
		Reflective(1);
		SpecExponent(2);
		Transparent(0);
		
		Sphere(50);

		translate(100, 0, 0);

		// Left Sphere
		Ambient(0.3, 0.3, 0.2);
		Diffuse(0.15164, 0.30648, 0.22648);
		Specular(0.628281, 0.555802, 0.366065);
    
		Reflective(1);
		SpecExponent(0.5);
		Transparent(0);
		
		Sphere(50);

		translate(-100, 0, 0);

		break;

	case 1:

		Background(0, 0, 0);

		AddLight(1.0, 0, -250, -200);		

		// Gold Cones
		Ambient(0.24725, 0.1995, 0.0745);
		Diffuse(0.75164, 0.60648, 0.22648);
		Specular(0.628281, 0.555802, 0.366065);
    
		Reflective(1);
		SpecExponent(89.2);
		Transparent(0);
		
		Cone (50,150);
		rotateX(140);
		translate(-70,30,0);

		Ambient(0.23125 , 0.23125 , 0.23125 );
		Diffuse(0.2775, 0.2775, 0.2775);
		Specular(0.773911, 0.773911, 0.773911);

		Reflective(1);
		SpecExponent(89.6);
		Transparent(0);		

		Cone (50,150);
		rotateX(140);
		translate(70,30,0);

		break;

	case 2:

		Background(0, 0, 0);

		AddLight(0.6, -100, -200, 0);
		AddLight(0.6, 100, -200, 0);

		/*
		// Silver
		Ambient(0.23125 , 0.23125 , 0.23125 );
		Diffuse(0.2775, 0.2775, 0.2775);
		Specular(0.773911, 0.773911, 0.773911);

		Reflective(1);
		SpecExponent(89.6);
		Transparent(0);
		*/
		/*
		// Brass
		Ambient(0.329412, 0.223529 , 0.027451 );
		Diffuse(0.780392, 0.568627, 0.113725);
		Specular(0.992157, 0.941176, 0.807843);

		Reflective(1);
		SpecExponent(27.8974);
		Transparent(0);

		*/

		Ambient(0.1, 0.1, 0.4);
		Diffuse(0.1, 0.1, 0.8);
		Specular(1, 1, 1);

		Reflective(1);
		SpecExponent(12);
		Transparent(0);

		Ambient(0.23125 , 0.23125 , 0.23125 );
		Diffuse(0.2775, 0.2775, 0.2775);
		Specular(0.773911, 0.773911, 0.773911);

		Reflective(1);
		SpecExponent(89.6);
		Transparent(0);
		
		// generate set of points for the plane
		Vec4 *P3;
		P3 = new Vec4[4];

		P3[0].x = -200;
		P3[0].y = -200;
		P3[0].z = 0;
		P3[0].w = 1;

		P3[1].x = -200;
		P3[1].y = 200;
		P3[1].z = 0;
		P3[1].w = 1;

		P3[2].x = 200;
		P3[2].y = 200;
		P3[2].z = 0;
		P3[2].w = 1;

		P3[3].x = 200;
		P3[3].y = -200;
		P3[3].z = 0;
		P3[3].w = 1;

		Polygon(4, P3);

		rotateX(-80);
		translate(0, 70, 100);

		// Use Gold for our surface properties
		Ambient(0.24725, 0.1995, 0.0745);
		Diffuse(0.75164, 0.60648, 0.22648);
		Specular(0.628281, 0.555802, 0.366065);
    
		Reflective(1);
		SpecExponent(89.2);
		Transparent(0);

		// front sphere
		Sphere(40);
		ObjList.translate(0, 50, -30);

		// left sphere
		Sphere(40);
		ObjList.translate(-40, 30, 20);

		// right sphere
		Sphere(40);
		ObjList.translate(40, 30, 20);

		// top sphere
		Sphere(40);
		ObjList.translate(0, -20, 0);

		// Right Cylinder
		Cylinder(20, 200);
		ObjList.translate(150,0,0);

		// Left Cylinder
		Cylinder(20, 200);
		ObjList.translate(-150,0,0);

		// transparent plane in front
		Vec4 *P4;
		P4 = new Vec4[4];

		P4[0].x = 0;
		P4[0].y = -1;
		P4[0].z = 0;
		P4[0].w = 1;

		P4[1].x = -2;
		P4[1].y = 0;
		P4[1].z = 0;
		P4[1].w = 1;

		P4[2].x = 2;
		P4[2].y = 0;
		P4[2].z = 0;
		P4[2].w = 1;

		Ambient(0.23125 , 0.23125 , 0.23125 );
		Diffuse(0.2775, 0.2775, 0.2775);
		Specular(0.773911, 0.773911, 0.773911);

		Reflective(1);
		SpecExponent(89.6);
		Transparent(1);

		Transparent(1);
		Polygon(3, P4);
		translate(45, 0,-150);
		scale(25,50,50);

		break;
    }
}



